#include "ch.h"
#include "hal.h"
#include "stm32f4_discovery_audio_codec.h"

/* Mask for the bit EN of the I2S CFGR register */
#define I2S_ENABLE_MASK                 0x0400
#define I2SCFGR_CLEAR_MASK        ((uint16_t)0xF040)

/* Codec audio Standards */
#define  CODEC_STANDARD                0x04
#define I2S_STANDARD                   I2S_Standard_Phillips         

/* The 7 bits Codec address (sent through I2C interface) */
#define CODEC_ADDRESS                   0x4A  /* b00100111 */

uint32_t Codec_Init(uint16_t OutputDevice, uint8_t Volume, uint32_t AudioFreq);
static uint32_t Codec_VolumeCtrl(uint8_t Volume);
/* Low layer codec functions */
static void     Codec_CtrlInterface_Init(void);
void     Codec_AudioInterface_Init(uint32_t AudioFreq);
static void     Codec_Reset(void);
static uint32_t Codec_WriteRegister(uint8_t RegisterAddr, uint8_t RegisterValue);
void     Codec_GPIO_Init(void);
/* Peripherals configuration functions */
static void     Audio_MAL_Init(void);

uint32_t EVAL_AUDIO_Init(uint16_t OutputDevice, uint8_t Volume, uint32_t AudioFreq)
{    
  /* Perform low layer Codec initialization */
  if (Codec_Init(OutputDevice, VOLUME_CONVERT(Volume), AudioFreq) != 0)
  {
    return 1;                
  }
  else
  {    
    /* I2S data transfer preparation:
    Prepare the Media to be used for the audio transfer from memory to I2S peripheral */
    Audio_MAL_Init();
    
    /* Return 0 when all operations are OK */
    return 0;
  }
}

uint32_t EVAL_AUDIO_VolumeCtl(uint8_t Volume)
{
  /* Call the codec volume control function with converted volume value */
  return (Codec_VolumeCtrl(VOLUME_CONVERT(Volume)));
}

extern Thread *MadThd;

static void I2SDmaTxInterrupt(SPIDriver* spip, uint32_t flags)

{
  if(spip == &SPID3)
  {
    if(flags & STM32_DMA_ISR_TCIF)
    {
      // let the decoder thread play the next samples
      chSysLockFromIsr();
      chEvtSignalI(MadThd, (eventmask_t)1);
      chSysUnlockFromIsr();
    }
  }
}

uint32_t Codec_Init(uint16_t OutputDevice, uint8_t Volume, uint32_t AudioFreq)
{
  uint32_t counter = 0; 

  /* Configure the Codec related IOs */
  Codec_GPIO_Init();   
  
  /* Reset the Codec Registers */
  Codec_Reset();

  /* Initialize the Control interface of the Audio Codec */
  Codec_CtrlInterface_Init();     
  
  /* Keep Codec powered OFF */
  counter += Codec_WriteRegister(0x02, 0x01);  
      
  counter += Codec_WriteRegister(0x04, 0xAF); /* SPK always OFF & HP always ON */
  
  /* Clock configuration: Auto detection */  
  counter += Codec_WriteRegister(0x05, 0x81);
  
  /* Set the Slave Mode and the audio Standard */  
  counter += Codec_WriteRegister(0x06, CODEC_STANDARD);
      
  /* Set the Master volume */
  Codec_VolumeCtrl(Volume);
  
  /* Power on the Codec */
  counter += Codec_WriteRegister(0x02, 0x9E);  
  
  /* Additional configuration for the CODEC. These configurations are done to reduce
      the time needed for the Codec to power off. If these configurations are removed, 
      then a long delay should be added between powering off the Codec and switching 
      off the I2S peripheral MCLK clock (which is the operating clock for Codec).
      If this delay is not inserted, then the codec will not shut down properly and
      it results in high noise after shut down. */
  
  /* Disable the analog soft ramp */
  counter += Codec_WriteRegister(0x0A, 0x00);
  /* Disable the limiter attack level */
  counter += Codec_WriteRegister(0x27, 0x00);
  /* Adjust Bass and Treble levels */
  counter += Codec_WriteRegister(0x1F, 0x0F);
  /* Adjust PCM volume level */
  counter += Codec_WriteRegister(0x1A, 0x0A);
  counter += Codec_WriteRegister(0x1B, 0x0A);

  /* Configure the I2S peripheral */
  Codec_AudioInterface_Init(AudioFreq);  
  
  /* Return communication control value */
  return counter;  
}

static uint32_t Codec_VolumeCtrl(uint8_t Volume)
{
  uint32_t counter = 0;
  
  if (Volume > 0xE6)
  {
    /* Set the Master volume */
    counter += Codec_WriteRegister(0x20, Volume - 0xE7); 
    counter += Codec_WriteRegister(0x21, Volume - 0xE7);     
  }
  else
  {
    /* Set the Master volume */
    counter += Codec_WriteRegister(0x20, Volume + 0x19); 
    counter += Codec_WriteRegister(0x21, Volume + 0x19); 
  }

  return counter;  
}

static void Codec_Reset(void)
{
  /* Power Down the codec */
  palClearPad(GPIOD, 4); // GPIO_WriteBit(AUDIO_RESET_GPIO, AUDIO_RESET_PIN, Bit_RESET);
  /* wait for a delay to insure registers erasing */
  chThdSleepMilliseconds(10);
  /* Power on the codec */
  palSetPad(GPIOD, 4); // GPIO_WriteBit(AUDIO_RESET_GPIO, AUDIO_RESET_PIN, Bit_SET);
}

static uint32_t Codec_WriteRegister(uint8_t RegisterAddr, uint8_t RegisterValue)
{
  uint32_t result = 0;
  uint8_t  txbuf[2];

  txbuf[0]=RegisterAddr; txbuf[1]=RegisterValue;
  result=i2cMasterTransmitTimeout(&I2CD1, CODEC_ADDRESS, txbuf, 2 , NULL, 0, 1000);

  /* Return the verifying value: 0 (Passed) or any other (Failed) */
  return result;  
}

static I2CConfig i2cfg = {
    OPMODE_I2C,
    400000,
    FAST_DUTY_CYCLE_2
    //STD_DUTY_CYCLE
};

static void Codec_CtrlInterface_Init(void)
{
  i2cStart(&I2CD1, &i2cfg);
}

#define I2S_Mode_MasterTx                ((uint16_t)0x0200)
#define I2S_Standard_Phillips           ((uint16_t)0x0000)
#define I2S_DataFormat_16b              ((uint16_t)0x0000)
#define I2S_MCLKOutput_Enable           ((uint16_t)0x0200)
#define I2S_AudioFreq_Default            ((uint32_t)2)
#define I2S_CPOL_Low                    ((uint16_t)0x0000)

void Codec_AudioInterface_Init(uint32_t AudioFreq)
{
  uint16_t tmpreg = 0, i2sdiv = 2, i2sodd = 0, packetlength = 1;
  uint32_t tmp = 0, i2sclk = 0;
  uint32_t pllm = 0, plln = 0, pllr = 0;

  rccEnableSPI3(FALSE);

/*----------------------- SPIx I2SCFGR & I2SPR Configuration -----------------*/
  /* Clear I2SMOD, I2SE, I2SCFG, PCMSYNC, I2SSTD, CKPOL, DATLEN and CHLEN bits */

  SPI3->I2SCFGR &= I2SCFGR_CLEAR_MASK;
  SPI3->I2SPR = 0x0002;

  /* Get the I2SCFGR register value */
  tmpreg = SPI3->I2SCFGR;

  /* If the default value has to be written, reinitialize i2sdiv and i2sodd*/

  if(AudioFreq == I2S_AudioFreq_Default)
  {
    i2sodd = (uint16_t)0;
    i2sdiv = (uint16_t)2;
  }
  /* If the requested audio frequency is not the default, compute the prescaler */
  else
  {

    /* Check the frame length (For the Prescaler computing) *******************/
      /* Packet length is 16 bits */
      packetlength = 1;

    /* Get I2S source Clock frequency  ****************************************/
    /* Set PLLI2S as I2S clock source */

    if ((RCC->CFGR & RCC_CFGR_I2SSRC) != 0)
    {
      RCC->CFGR &= ~(uint32_t)RCC_CFGR_I2SSRC;
    }

    //chprintf((BaseChannel *)&SD,"CFGI2SR : %x\n\r", RCC->PLLI2SCFGR);

    /* Get the PLLI2SN value */
    plln = (RCC->PLLI2SCFGR & RCC_PLLI2SCFGR_PLLI2SN) >> 6;
    /* Get the PLLI2SR value */
    pllr = (RCC->PLLI2SCFGR & RCC_PLLI2SCFGR_PLLI2SR) >> 28;

    /* Get the PLLM value */
    pllm = (uint32_t)(RCC->PLLCFGR & RCC_PLLCFGR_PLLM);

    /* Get the I2S source clock value */
    i2sclk = (uint32_t)(((HSE_VALUE / pllm) * plln) / pllr);

    //chprintf((BaseChannel *)&SD,"PLL : N=%d, R=%d, M=%d, I2SCLK=%d\n\r", plln,pllr,pllm, i2sclk);

    /* Compute the Real divider depending on the MCLK output state, with a floating point */
      /* MCLK output is enabled */

    tmp = (uint16_t)(((((i2sclk / 256) * 10) / AudioFreq)) + 5);

    /* Remove the flatting point */
    tmp = tmp / 10;

    /* Check the parity of the divider */
    i2sodd = (uint16_t)(tmp & (uint16_t)0x0001);

    /* Compute the i2sdiv prescaler */
    i2sdiv = (uint16_t)((tmp - i2sodd) / 2);

    /* Get the Mask for the Odd bit (SPI_I2SPR[8]) register */
    i2sodd = (uint16_t) (i2sodd << 8);
  }

  //chprintf((BaseChannel *)&SD,"DIVS I2S CLK : div=%d, odd=%d\n\r", i2sdiv, i2sodd);

  /* Test if the divider is 1 or 0 or greater than 0xFF */
  if ((i2sdiv < 2) || (i2sdiv > 0xFF))
  {
    /* Set the default values */
    i2sdiv = 2;
    i2sodd = 0;
  }

  /* Write to SPIx I2SPR register the computed value */
  SPI3->I2SPR = (uint16_t)((uint16_t)i2sdiv | (uint16_t)(i2sodd | I2S_MCLKOutput_Enable));

  /* Configure the I2S with the SPI_InitStruct values */
  tmpreg |= SPI_I2SCFGR_I2SMOD | I2S_Mode_MasterTx | \
                  I2S_Standard_Phillips | I2S_DataFormat_16b | \
                  I2S_CPOL_Low | \
                  SPI_I2SCFGR_I2SE;

  /* Write to SPIx I2SCFGR */
  SPI3->I2SCFGR = tmpreg;
}

void Codec_GPIO_Init(void)
{
  /* Audio reset pin configuration -------------------------------------------------*/
  palSetPadMode(GPIOD, 4, PAL_MODE_OUTPUT_PUSHPULL | PAL_STM32_OSPEED_MID2);						// Audio RESET = PD4

  /* CODEC_I2C SCL and SDA pins configuration -------------------------------------*/

  palSetPadMode(GPIOB, 6, PAL_STM32_OTYPE_OPENDRAIN|PAL_MODE_ALTERNATE(4) | PAL_STM32_OSPEED_MID2);		//SCL	= PB6	-> I2CD1
  palSetPadMode(GPIOB, 9, PAL_STM32_OTYPE_OPENDRAIN|PAL_MODE_ALTERNATE(4) | PAL_STM32_OSPEED_MID2);		//SDA	= PB9

  /* CODEC_I2S pins configuration: WS, SCK and SD pins -----------------------------*/

  palSetPadMode(GPIOC, 10, PAL_STM32_OTYPE_PUSHPULL | PAL_MODE_ALTERNATE(6) | PAL_STM32_OSPEED_MID2);                //SCK   = PC10  -> SPI3
  palSetPadMode(GPIOC, 12, PAL_STM32_OTYPE_PUSHPULL | PAL_MODE_ALTERNATE(6) | PAL_STM32_OSPEED_MID2);                //SD    = PC12
  palSetPadMode(GPIOA,  4, PAL_STM32_OTYPE_PUSHPULL | PAL_MODE_ALTERNATE(6) | PAL_STM32_OSPEED_MID2);                //WS    = PA4
  palSetPadMode(GPIOC,  7, PAL_STM32_OTYPE_PUSHPULL | PAL_MODE_ALTERNATE(6) | PAL_STM32_OSPEED_MID2);                //MCK   = PC7
}

static void Audio_MAL_Init(void)  
{ 
    dmaStreamAllocate(SPID3.dmatx,
                    STM32_SPI_SPI3_IRQ_PRIORITY,
                    (stm32_dmaisr_t)I2SDmaTxInterrupt,
                    &SPID3);

    // Default SPID3.tx Config -> CHANNEL / PRIORITY / M2P / DMEIE / TEIE

    rccEnableSPI3(FALSE);
    SPID3.txdmamode |= ( STM32_DMA_CR_PSIZE_HWORD | 0x30000 |  
                         STM32_DMA_CR_MINC | STM32_DMA_CR_MSIZE_HWORD | 
                         STM32_DMA_CR_DBM | STM32_DMA_CR_TCIE );
    dmaStreamSetMode(SPID3.dmatx, SPID3.txdmamode );
    dmaStreamSetFIFO(SPID3.dmatx, STM32_DMA_FCR_DMDIS | STM32_DMA_FCR_FTH_HALF );
    dmaStreamSetPeripheral(SPID3.dmatx, &SPID3.spi->DR);

    SPID3.spi->CR2 |= SPI_CR2_TXDMAEN;
}

void Audio_MAL_Play(uint32_t Addr0, uint32_t Addr1, uint32_t Size)
{         
  dmaStreamSetMemory0(SPID3.dmatx, Addr0);
  dmaStreamSetMemory1(SPID3.dmatx, Addr1);
  dmaStreamSetTransactionSize(SPID3.dmatx, Size/2);
  dmaStreamEnable(SPID3.dmatx);
}

